import signal
import asyncio
import urllib.parse
import urllib.request
import json

from .core import Miner


def search(query=None, params=None, callback=None, mine_ids=None, info_only=None,
           **kwargs):
    query = '(*:*)' if not query else query
    params = params if params else {}
    mine_ids = True if mine_ids else False
    info_only = True if info_only else info_only

    if not kwargs.get('loop'):
        loop = asyncio.get_event_loop()
    else:
        loop = kwargs['loop']
    miner = Miner(loop, **kwargs)

    if info_only:
        url = miner.make_url('/advancedsearch.php?')
        params = urllib.parse.urlencode(miner.get_search_params(query, params))
        resp = urllib.request.urlopen(url + params)
        j = json.loads(resp.read().decode('utf-8'))
        search_info = j['responseHeader']
        search_info['numFound'] = j.get('response', {}).get('numFound', 0)
        return search_info

    try:
        loop.add_signal_handler(signal.SIGINT, loop.stop)
        loop.run_until_complete(miner.search(query, params=params, callback=callback,
                                             mine_ids=mine_ids))
    except RuntimeError:
        pass


def mine_items(identifiers, params=None, callback=None, **kwargs):
    """Concurrently retrieve metadata from Archive.org items.

    Args:
        identifiers: An iterable yielding Archive.org item identifiers.
        response_callback: A callback function to call on each
            aiohttp.client.ClientResponse object successfully returned.
        max_tasks: The maximum number of tasks to run concurrently.
        max_retries: The maximum number of times to retry a given request.
        secure: A boolean indicating if a secure connection should be used or not.
        hosts: A list of hosts to cycle through randomly for each request.
    """
    if not kwargs.get('loop'):
        loop = asyncio.get_event_loop()
    else:
        loop = kwargs['loop']
    miner = Miner(loop, **kwargs)
    try:
        loop.add_signal_handler(signal.SIGINT, loop.stop)
        loop.run_until_complete(miner.mine_items(identifiers, callback))
    except RuntimeError:
        pass



#!/usr/bin/env python3
"""Concurrently retrieve metadata from Archive.org items.

usage: ia-mine (<itemlist> | -) [--workers WORKERS] [--cache]
               [--retries RETRIES] [--secure] [--hosts HOSTS]
       ia-mine --search QUERY [--field FIELD... | --mine-ids | --itemlist]
               [--workers WORKERS] [--cache] [--retries RETRIES]
               [--secure] [--hosts HOSTS]
       ia-mine [-h | --version] 


positional arguments:
  itemlist              A file containing Archive.org identifiers, one per
                        line, for which to retrieve metadata from. If no
                        itemlist is provided, identifiers will be read from
                        stdin.

optional arguments:
  -h, --help            show this help message and exit
  -v, --version         show program's version number and exit
  -s, --search QUERY    mine search search results.
  -m, --mine-ids        mine items returned from search results.
                        [default: False]
  -f, --field FIELD     fields to include in search results.
                        [default: identifier]
  -i, --itemlist        dump identifiers only to stdout. [default: False]
  -w, --workers WORKERS
                        The maximum number of tasks to run at once.
                        [default: 100]
  -c, --cache           Cache item metadata on Archive.org. Items are not
                        cached are not cached by default.
  -r, --retries RETRIES
                        The maximum number of retries for each item.
                        [default: 10]
  -s, --secure          Use HTTPS. HTTP is used by default.
  -H, --hosts HOSTS     A file containing a list of hosts to shuffle through.

"""
def suppress_keyboard_interrupt_message():
    """Register a new excepthook to suppress KeyboardInterrupt
    exception messages, and exit with status code 130.

    """
    old_excepthook = sys.excepthook

    def new_hook(type, value, traceback):
        if type != KeyboardInterrupt:
            old_excepthook(type, value, traceback)
        else:
            sys.exit(130)

    sys.excepthook = new_hook

import os
import sys
suppress_keyboard_interrupt_message()

from docopt import docopt, DocoptExit
from schema import Schema, Use, Or, SchemaError

from .api import mine_items, search
from ._version import __version__


def print_itemlist(resp):
    j = yield from resp.json()
    for doc in j.get('response', {}).get('docs', []):
        print(doc.get('identifier'))

def main(argv=None):
    # If ia-wrapper calls main with argv argument, strip the
    # "mine" subcommand from args. 
    argv = argv[1:] if argv else sys.argv[1:]

    # Catch DocoptExit error and write to stderr manually.
    # Otherwise error's vanish if executed from a pex binary. 
    try:
        args = docopt(__doc__, version=__version__, argv=argv, help=True)
    except DocoptExit as exc:
        sys.exit(sys.stderr.write('{}\n'.format(exc.code)))

    # Validate args.
    open_file_or_stdin = lambda f: sys.stdin if (f == '-') or (not f) else open(f)
    parse_hosts = lambda f: [x.strip() for x in open(f) if x.strip()]
    schema = Schema({object: bool,
        '--search': Or(None, Use(str)),
        '--field': list,
        '--hosts': Or(None, 
            Use(parse_hosts, error='--hosts=HOSTS should be a readable file.')),
        '--retries': Use(int, 'RETRIES should be an integer.'),
        '<itemlist>': Use(open_file_or_stdin, error='<itemlist> should be readable'),
        '--workers': Use(int, error='--workers=WORKERS should be an integer.'),
    })
    try:
        args = schema.validate(args)
    except SchemaError as exc:
        sys.exit(sys.stderr.write('{0}error: {1}\n'.format(__doc__, str(exc))))

    if args['--search']:
        callback = print_itemlist if args['--itemlist'] else None
        params = {}
        if 'identifier' not in args['--field']:
            args['--field'].append('identifier')
        for i, f in enumerate(args['--field']):
            params['fl[{}]'.format(i)] = f
        search(args['--search'],
               params=params,
               callback=callback,
               mine_ids=args['--mine-ids'],
               max_tasks=args['--workers'],
               retries=args['--retries'],
               secure=args['--secure'],
               hosts=args['--hosts'])
        sys.exit(0)
    else:
        # Exit with 2 if stdin appears to be empty.
        if args['-']:
            if (not os.fstat(sys.stdin.fileno()).st_size > 0) and (sys.stdin.seekable()):
                sys.exit(2)

        mine_items(args['<itemlist>'],
                   max_tasks=args['--workers'],
                   retries=args['--retries'],
                   secure=args['--secure'],
                   hosts=args['--hosts'])



import asyncio
import functools
from copy import deepcopy
import sys
import random

import aiohttp


class Miner:

    def __init__(self, loop=None, max_tasks=None, retries=None, secure=None, hosts=None,
                 params=None):
        # Set default values for kwargs.
        max_tasks = 100 if not max_tasks else max_tasks
        max_retries = 10 if not retries else retries
        protocol = 'http://' if not secure else 'https://'
        loop = asyncio.get_event_loop() if not loop else loop

        self.protocol = protocol
        self.hosts = hosts
        self.max_retries = max_retries

        # Asyncio/Aiohttp settings.
        self.tasks = set()
        self.loop = loop
        self.sem = asyncio.Semaphore(max_tasks)
        self.connector = aiohttp.TCPConnector(share_cookies=True, loop=loop)

    @asyncio.coroutine
    def mine_items(self, identifiers, params=None, callback=None):
        callback = functools.partial(self.handle_response, callback=callback)
        # By default, don't cache item metadata in redis.
        params = {'dontcache': 1} if not params else {}

        def _md_request_generator():
            for identifier in identifiers:
                url = self.make_url('/metadata/{}'.format(identifier))
                yield (url, params, callback)

        task = asyncio.Task(self.add_requests(_md_request_generator()))
        task.add_done_callback(self.tasks.remove)
        self.tasks.add(task)

        # Sleep until all tasks are complete.
        while self.tasks:
            yield from asyncio.sleep(.1)

    @asyncio.coroutine
    def handle_response(self, resp, callback=None):
        if callback:
            yield from callback(resp)
            resp.close()
        else:
            content = yield from resp.read()
            resp.close()
            print(content.decode('utf-8'))

    @asyncio.coroutine
    def search(self, query=None, params=None, callback=None, mine_ids=None):
        search_params = self.get_search_params(query, params)
        url = self.make_url(path='/advancedsearch.php')

        search_info = yield from self.get_search_info(url, search_params)
        total_results = search_info.get('response', {}).get('numFound', 0)
        total_pages = (int(total_results/search_params['rows']))

        for page in range(1, (total_pages + 2)):
            params = deepcopy(search_params)
            params['page'] = page
            cb = functools.partial(self._handle_search_results, callback=callback,
                                   mine_ids=mine_ids)
            task = asyncio.Task(self.add_requests([(url, params, cb)]))
            task.add_done_callback(self.tasks.remove)
            self.tasks.add(task)

        # Sleep until all tasks are complete.
        while self.tasks:
            yield from asyncio.sleep(.1)

    def get_search_params(self, query, params):
        default_rows = 1000
        search_params = {
            'q': '(*:*)',
            'fl[]': 'identifier',
            'page': 1,
            'output': 'json',
        }
        if params:
            search_params.update({k: v for k, v in params.items() if v})
        if query:
            search_params['q'] = query
        if 'rows' not in search_params:
            search_params['rows'] = default_rows
        return search_params

    def get_search_info(self, url, params):
        p = deepcopy(params)
        p['rows'] = 0
        resp = yield from aiohttp.request('get', url, params=p)
        resp.close()
        return (yield from resp.json())

    @asyncio.coroutine
    def _handle_search_results(self, resp, mine_ids=None, params=None, callback=None):
        params = {'dontcache': 1} if not params else params
        callback = functools.partial(self.handle_response, callback=callback)

        if not mine_ids:
            task = asyncio.Task(self.handle_response(resp, callback=callback))
            task.add_done_callback(self.tasks.remove)
            self.tasks.add(task)
        else:
            j = yield from resp.json()
            resp.close()
            requests = []
            for doc in j.get('response', {}).get('docs', []):
                url = self.make_url('/metadata/{}'.format(doc['identifier']))
                requests.append((url, params, callback))
            if requests:
                task = asyncio.Task(self.add_requests(requests))
                task.add_done_callback(self.tasks.remove)
                self.tasks.add(task)

    @asyncio.coroutine
    def add_requests(self, requests):
        for url, params, callback in requests:
            yield from self.sem.acquire()
            task = asyncio.Task(self.make_request(url, params, callback))
            task.add_done_callback(lambda t: self.sem.release())
            task.add_done_callback(self.tasks.remove)
            self.tasks.add(task)

    @asyncio.coroutine
    def make_request(self, url, params=None, callback=None):
        params = {'dontcache': 1} if not params else params
        retry = 0
        while True:
            retry += 1
            try:
                resp = yield from aiohttp.request(
                    'get', url, params=params, connector=self.connector)
                if callback:
                    return (yield from callback(resp))
                else:
                    return resp
            except Exception as exc:
                raise exc
                sys.stderr.write('{0} has error {1}\n'.format(url, repr(exc)))
                if retry >= self.max_retries:
                    return
                sys.stderr.write(
                    '... retry {0}/{1} {2}\n'.format(retry, self.max_retries, url))
                yield from asyncio.sleep(1)

    def make_url(self, path):
        if self.hosts:
            host = self.hosts[random.randrange(len(self.hosts))]
        else:
            host = 'archive.org'
        return self.protocol + host + path



#!/usr/bin/env python3
"""
IA Mine
~~~~~~~

Concurrently retrieve metadata from Archive.org items.

:copyright: (c) 2015 by Internet Archive.
:license: AGPL 3, see LICENSE for more details.

"""
__title__ = 'iamine'
__author__ = 'Jacob M. Johnson'
__license__ = 'AGPL 3'
__copyright__ = 'Copyright 2015 Internet Archive'

from ._version import __version__
from .core import Miner
from .api import mine_items, search

__all__ = ['Miner', 'mine_items', 'search', '__version__']
